"""
Models for RabbitMQ objects.
"""
import datetime
import json
import pydantic

from enum import Enum
from typing import Any, Generic, Type, TypeVar, Optional, List, Union, Dict
from pydantic import AnyUrl

TRabbitObject = TypeVar('TRabbitObject', bound='RabbitObject')


def alias_replace_underscore(field_name):
    """Replace '_' with '-' in field name

    Some field names contain '-' which is interpreted as an operator by python, and can
    therefore not be used in field names defined on our pydantic models. Using this as
    an alias_generator we are still able to parse them.
    """
    return field_name.replace('_', '-')


class NonEmptyStr(str):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if not isinstance(v, str):
            raise ValueError(f'str expected, not {type(v)}')
        if not v:
            raise ValueError('non-empty string expected')
        return v


class RabbitObject(Generic[TRabbitObject], pydantic.BaseModel):
    """
    Abstract RabbitMQ object.
    """

    @classmethod
    def from_dict(cls, data):
        return cls(**data)

    @classmethod
    def from_json(cls, json_data):
        data = json.loads(json_data)
        return cls.from_dict(data)


class User(RabbitObject):
    name: str
    password_hash: str
    hashing_algorithm: str
    tags: str


class Metadata(RabbitObject):
    description: str
    tags: List[NonEmptyStr]


class Arguments(RabbitObject):
    x_queue_type: Optional[NonEmptyStr]

    class Config(RabbitObject.Config):
        alias_generator = alias_replace_underscore
        allow_population_by_field_name = True


class Vhost(RabbitObject):
    name: NonEmptyStr
    description: str
    tags: List[NonEmptyStr]
    metadata: Metadata
    tracing: bool
    cluster_state: Dict[NonEmptyStr, NonEmptyStr]


class ShovelValue(RabbitObject):
    ack_mode: NonEmptyStr
    add_forward_headers: bool
    dest_add_forward_headers: Optional[bool]
    dest_exchange: Optional[NonEmptyStr]
    dest_queue: Optional[NonEmptyStr]
    dest_exchange_key: Optional[NonEmptyStr]
    dest_protocol: NonEmptyStr
    dest_uri: NonEmptyStr
    delete_after: NonEmptyStr
    src_delete_after: Optional[NonEmptyStr]
    src_protocol: NonEmptyStr
    src_queue: Optional[NonEmptyStr]
    src_exchange: Optional[NonEmptyStr]
    src_uri: NonEmptyStr

    class Config(RabbitObject.Config):
        alias_generator = alias_replace_underscore
        allow_population_by_field_name = True


class Shovel(RabbitObject):
    value: ShovelValue
    vhost: NonEmptyStr
    component: NonEmptyStr
    name: NonEmptyStr


class ShovelStatus(RabbitObject):
    node: NonEmptyStr
    timestamp: datetime.datetime
    name: NonEmptyStr
    vhost: NonEmptyStr
    type: NonEmptyStr
    state: NonEmptyStr
    # Only exists when the shovel is in running state
    src_uri: Optional[NonEmptyStr]
    src_protocol: Optional[NonEmptyStr]
    dest_protocol: Optional[NonEmptyStr]
    dest_uri: Optional[NonEmptyStr]
    # Only one of each (src_queue, src_exchange and dest_queue, dest_exchange) will be defined
    src_queue: Optional[NonEmptyStr]
    src_exchange: Optional[NonEmptyStr]
    dest_queue: Optional[NonEmptyStr]
    dest_exchange: Optional[NonEmptyStr]

    @pydantic.validator('timestamp', pre=True)
    def parse_timestamp(cls, value):
        """Parse timestamp on format YYYY-MM-DD HH:MM:SS or YYYY-MM-DD H:MM:SS"""
        return datetime.datetime.strptime(value, "%Y-%m-%d %H:%M:%S")


class Permission(RabbitObject):
    vhost: NonEmptyStr
    user: NonEmptyStr
    configure: NonEmptyStr
    write: NonEmptyStr
    read: NonEmptyStr


class Exchange(RabbitObject):
    arguments: Arguments
    auto_delete: bool
    durable: bool
    internal: bool
    name: str
    type: NonEmptyStr
    user_who_performed_action: NonEmptyStr
    vhost: NonEmptyStr


class Queue(RabbitObject):
    arguments: Arguments
    auto_delete: bool
    consumers: Optional[int]
    durable: bool
    messages: Optional[int]
    name: str
    node: NonEmptyStr
    state: Optional[NonEmptyStr]
    type: NonEmptyStr
    vhost: NonEmptyStr


class ConnectionDetails(RabbitObject):
    name: str
    peer_host: Optional[NonEmptyStr]


class Connection(ConnectionDetails):
    auth_mechanism: Optional[NonEmptyStr]
    channel_max: Optional[int]
    channels: Optional[int]
    client_properties: dict
    host: Optional[NonEmptyStr]
    node: NonEmptyStr
    peer_cert_issuer: Optional[NonEmptyStr]
    peer_cert_subject: Optional[NonEmptyStr]
    peer_cert_validity: Optional[NonEmptyStr]
    peer_port: Optional[int]
    port: Optional[int]
    ssl: Optional[bool]
    ssl_cipher: Optional[NonEmptyStr]
    ssl_hash: Optional[NonEmptyStr]
    ssl_key_exchange: Optional[NonEmptyStr]
    ssl_protocol: Optional[NonEmptyStr]
    state: Optional[NonEmptyStr]
    timeout: Optional[int]
    type: NonEmptyStr
    user: NonEmptyStr
    user_who_performed_action: NonEmptyStr


class Channel(RabbitObject):
    acks_uncommitted: int
    confirm: bool
    connection_details: ConnectionDetails
    consumer_count: int
    name: str
    node: NonEmptyStr
    state: NonEmptyStr
    user: NonEmptyStr
    user_who_performed_action: NonEmptyStr
    vhost: NonEmptyStr


class Binding(RabbitObject):
    source: str
    vhost: str
    destination: NonEmptyStr
    destination_type: NonEmptyStr
    routing_key: str
    arguments: Arguments
    properties_key: NonEmptyStr


class Definitions(RabbitObject):
    rabbit_version: NonEmptyStr
    rabbitmq_version: NonEmptyStr
    product_name: NonEmptyStr
    product_version: NonEmptyStr
    users: List[User]
    permissions: List[Permission]
    topic_permissions: list
    parameters: list
    global_parameters: list
    policies: list


class Policy(RabbitObject):
    vhost: NonEmptyStr
    name: NonEmptyStr
    pattern: NonEmptyStr
    apply_to: Optional[NonEmptyStr]
    definition: dict
    priority: Optional[int]

    class Config(RabbitObject.Config):
        alias_generator = alias_replace_underscore
        allow_population_by_field_name = True
